# Generated by Django 4.1.4 on 2022-12-23 01:44

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('password_policies', '0004_alter_passwordrecord_options'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='passwordrecord',
            options={'get_latest_by': 'date', 'ordering': ['-date']},
        ),
    ]



from django.apps import AppConfig


class PasswordPoliciesConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'password_policies'

    def ready(self):
        # Implicitly connect signal handlers decorated with @receiver.
        from . import signals  # noqa



import re
from datetime import timedelta

from django.core.exceptions import ValidationError
from django.contrib.auth.hashers import check_password
from django.utils.translation import gettext as _
from django.utils import timezone

from .models import PasswordRecord


class ComplexityValidator:
    def __init__(self, **kwargs):
        self.min_char_types = kwargs.pop('min_char_types', 4)
        self.min_chars_of_each_type = [
            ('min_numeric_chars', r'[0-9]'),
            ('min_uppercase_chars', r'[A-Z]'),
            ('min_lowercase_chars', r'[a-z]'),
            ('min_symbol_chars', r'[^0-9A-Za-z]'),
        ]
        for attr, _regex in self.min_chars_of_each_type:
            setattr(
                self, attr,
                kwargs.get(attr, 1)
            )

    def validate(self, password, user=None):
        password_valid = True
        errors = []
        char_types_contained = 0
        for type_, regex in self.min_chars_of_each_type:
            find = re.findall(regex, password)
            required = getattr(self, type_)
            if len(find) < required:
                password_valid = False
                errors.append(f"至少{required}個{type_}字元")
            if find:
                char_types_contained += 1

        if char_types_contained < self.min_char_types:
            password_valid = False
            errors.append(f"大寫、小寫、數字、特殊符號中至少{self.min_char_types}種字元")

        if not password_valid:
            raise ValidationError(
                f"密碼應包含{'；'.join(errors)}。",
                code='password_lacks_numeric_or_symbols',
            )

    def get_help_text(self):
        requirements = []
        for type_, regex in self.min_chars_of_each_type:
            required = getattr(self, type_)
            if required:
                requirements.append(f"至少{required}個{type_}字元")
        if self.min_char_types:
            requirements.append(
                f"大寫、小寫、數字、特殊符號中至少{self.min_char_types}種字元"
            )

        return f"密碼應包含{'；'.join(requirements)}。"


class RepeatedValidator:
    # 密碼hash方式，參考 django.contrib.auth.base_user.AbstractBaseUser
    # set_password(), check_password()
    # Validator寫法參考：
    # https://docs.djangoproject.com/en/4.1/topics/auth/passwords/#writing-your-own-validator

    def __init__(self, record_length=3):
        if record_length <= 0:
            raise ValueError('record_length must be larger than 0.')
        self.record_length = record_length

    def validate(self, password, user=None):
        # In case there is no user, this validator is not applicable.
        if user is None:
            return None

        stored_password_records = (
            PasswordRecord.objects.filter(user=user)
        )
        if not stored_password_records:
            return None
        for record in stored_password_records[:self.record_length]:
            if check_password(password, record.password):
                raise ValidationError(
                    self.get_help_text(),
                    code='password_repeated',
                )

    def get_help_text(self):
        return _(
            f"密碼不可與最近{self.record_length}次使用過的密碼重複。"
        )


class MinimumResetIntervalValidator:

    def __init__(self, min_interval_days=1):
        self.min_interval = timedelta(days=min_interval_days)

    def validate(self, password, user=None):
        # In case there is no user, this validator is not applicable.
        if user is None:
            return None
        try:
            latest_password_record = (
                PasswordRecord.objects.filter(user=user).latest()
            )
        except PasswordRecord.DoesNotExist:
            return None
        if (timezone.now() - latest_password_record.date) \
                < self.min_interval:
            raise ValidationError(
                _(f"距上次變更密碼須至少間隔{self.min_interval.days}日。"),
                code='password_reset_interval',
            )

    def get_help_text(self):
        return _(
            f"距上次變更密碼須至少間隔{self.min_interval.days}日。"
        )



from django.db.models.signals import post_save
from django.dispatch import receiver
from django.contrib.auth import get_user_model
from django.core.exceptions import ObjectDoesNotExist


@receiver(post_save, sender=get_user_model())
def create_password_record(sender, instance, **kwargs):
    try:
        latest_record = instance.password_records.latest()
        if latest_record.password == instance.password:
            return
    except ObjectDoesNotExist:
        pass
    instance.password_records.create(password=instance.password)



# Generated by Django 4.1.4 on 2022-12-19 19:54

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('password_policies', '0002_auto_20221215_1203'),
    ]

    operations = [
        migrations.AlterField(
            model_name='passwordrecord',
            name='user',
            field=models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, related_name='password_records', to=settings.AUTH_USER_MODEL),
        ),
    ]



# Generated by Django 4.0.7 on 2022-12-15 04:03

from django.db import migrations


class Migration(migrations.Migration):

    def create_password_record(apps, schema_editor):
        # We get the model from the versioned app registry;
        # if we directly import it, it'll be the wrong version
        User = apps.get_model("auth", "User")
        PasswordRecord = apps.get_model(
            "password_policies", "PasswordRecord"
        )

        for user in User.objects.all():
            PasswordRecord.objects.create(user=user, password=user.password)

    dependencies = [
        ('password_policies', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_password_record)
    ]



from django.db import models
from django.conf import settings


class PasswordRecord(models.Model):
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='password_records',
        on_delete=models.CASCADE,
        editable=False
    )

    # max_length 參考 django.contrib.auth.base_user.AbstractBaseUser.password
    password = models.CharField(
        'Password hash',
        max_length=128,
        editable=False
    )
    date = models.DateTimeField(
        'Date',
        auto_now_add=True,
        editable=False
    )

    class Meta:
        get_latest_by = 'date'
        ordering = ['-date']



from datetime import timedelta
from django.conf import settings
from django.shortcuts import redirect
from django.utils import timezone
from django.urls import resolve
from django.urls import reverse


class PasswordExpirationMiddleware:
    def __init__(self, get_response):
        # One-time configuration and initialization.
        self.get_response = get_response
        self.expiration_days = timedelta(
            days=getattr(settings, 'PASSWORD_EXPIRATION_DAYS', 90.0)
        )

    def __call__(self, request):
        # Code to be executed for each request before
        # the view (and later middleware) are called.

        resolve_match = resolve(request.path)
        if resolve_match.app_name == 'admin' and request.user.is_authenticated:
            if resolve_match.url_name != 'password_change':
                latest_record = request.user.password_records.latest()
                if (timezone.now() - latest_record.date) \
                        >= self.expiration_days:
                    return redirect(reverse(
                        "admin:password_change",
                        current_app=resolve_match.namespace
                    ))

        response = self.get_response(request)

        # Code to be executed for each request/response after
        # the view is called.

        return response



# Generated by Django 4.0.7 on 2022-12-15 03:27

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='PasswordRecord',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('password', models.CharField(editable=False, max_length=128, verbose_name='Password hash')),
                ('date', models.DateTimeField(auto_now_add=True, verbose_name='Date')),
                ('user', models.ForeignKey(editable=False, on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]
